from __future__ import annotations

from requests import Response


def format_return(function):
    """
    Decorator to evaluate the requests payload returned
    """

    def wrapped_answer(*args, **kwargs):
        """
        Decorator wrapper
        """
        req = function(*args, **kwargs)
        if isinstance(req, Response):
            try:
                return req.json()
            except Exception as error:
                print("request response not in JSON")
                print(error)
                return req.text
        return req

    return wrapped_answer



# generated by datamodel-codegen:
#   filename:  user-mappings.spec.json

from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, List, Optional, Union


@dataclass
class DetailedIdentity:
    username: str
    """
    Human friendly username used for ACLs/Interceptors
    """
    principal: Optional[str] = None
    """
    The Identity ID (i.e. claims.sub value when using OAuth)
    """
    groups: Optional[List[str]] = None
    """
    List of groups to be mapped to the principal. Groups apply to all of GW, not just the vClusters
    """


Identities = List[Union[str, DetailedIdentity]]


@dataclass
class UserMappings:
    """
    vCluster identities user-mappings.
    """

    identities: Identities


@dataclass
class UserMappingsConfig:
    """
    Specification for user-mapping management
    """

    userMappings: Optional[Dict[str, UserMappings]] = None



# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 John Mille <john@ews-network.net>

from __future__ import annotations


def set_plugings_actions_parsers(plugins_subparsers):
    list_parser = plugins_subparsers.add_parser(name="list", help="List all plugins")
    list_parser.add_argument("--extended", action="store_true", default=False)
    list_parser.add_argument(
        "--as-list", action="store_true", default=False, help="Returns only the list"
    )



# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 John Mille <john@ews-network.net>

from __future__ import annotations

import json
import warnings

import yaml

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader

from cdk_proxy_api_client.interceptors import Interceptors
from cdk_proxy_api_client.proxy_api import ProxyClient
from compose_x_common.compose_x_common import keyisset, set_else_none

from cdk_gw_tools.cli_tools import load_config_file


def interceptors_actions(proxy: ProxyClient, action: str, **kwargs):
    """Triggers function according to CLI Input"""

    _interceptor_client = Interceptors(proxy)
    if action == "list-all":
        return _interceptor_client.get_all_gw_interceptors().json()
    if action == "list":
        _interceptors = _interceptor_client.get_all_interceptor(
            is_global=kwargs.get("is_global"),
            username=kwargs.get("username"),
            group_name=kwargs.get("group_name"),
            vcluster_name=kwargs.get("vcluster_name"),
        ).json()
        if keyisset("IgnoreReadOnly", kwargs):
            return [
                _interceptor
                for _interceptor in _interceptors.values()
                if _interceptor["pluginClass"]
                != "io.conduktor.gateway.interceptor.safeguard.ReadOnlyTopicPolicyPlugin"
            ]
        return _interceptors
    elif action == "create-update":
        warnings.warn(
            "We highly recommend to use this command for testing only."
            "Use the import command to declare the interceptors you want to define as-code"
        )
        config = load_config_file(kwargs["config_file_path"])

        _req = _interceptor_client.update_interceptor(
            kwargs.get("interceptor_name"),
            config,
            is_global=kwargs.get("is_global"),
            vcluster_name=kwargs["vcluster_name"],
            username=kwargs.get("username"),
            group_name=kwargs.get("group_name"),
        ).json()
        return _req
    elif action == "delete":
        _req = _interceptor_client.delete_interceptor(
            interceptor_name=kwargs.get("interceptor_name"),
            is_global=kwargs.get("is_global"),
            vcluster_name=kwargs["vcluster_name"],
            username=kwargs.get("username"),
            group_name=kwargs.get("group_name"),
        ).status_code
        return _req
    elif action == "import-from-config":
        _loaded_config = load_config_file(kwargs["config_file_path"])
        if not keyisset("interceptors", _loaded_config):
            return
        from cdk_gw_tools.cli_tools.set_update_interceptors import (
            set_update_interceptors,
        )

        _req = set_update_interceptors(proxy, _loaded_config)
    return None



#   SPDX-License-Identifier: Apache-2.0
#   Copyright 2023 John Mille <john@ews-network.net>

from __future__ import annotations

import json
import logging

from boto3.session import Session
from compose_x_common.compose_x_common import keyisset

from cdk_gw_tools.aws.gw_handling import (
    get_cdk_gw_admin_creds,
    get_new_token_for_vcluster,
    new_gateway_vcluster_secret_value,
)

if logging.getLogger().hasHandlers():
    logging.getLogger().setLevel(logging.INFO)
else:
    logging.basicConfig(level=logging.INFO)

logger = logging.getLogger()


def lambda_handler(event, context):
    lambda_session = Session()
    if keyisset("vcluster", event) and keyisset("expiry", event):
        gateway_secret = get_cdk_gw_admin_creds(lambda_session)
        return {
            "token": get_new_token_for_vcluster(
                gateway_secret,
                {
                    "vcluster": event["vcluster"],
                    "username": event.get("username") or event["vcluster"],
                },
                int(event["expiry"]),
                token_only=True,
            )
        }
    else:
        secret_arn = event["SecretId"]
        token = event["ClientRequestToken"]
        step = event["Step"]

        current_value = lambda_session.client("secretsmanager").get_secret_value(
            SecretId=secret_arn, VersionStage="AWSCURRENT"
        )["SecretString"]
        if isinstance(current_value, str):
            try:
                current_value = json.loads(current_value)
                logger.info("Successfully decode JSON value from SecretString")
            except json.JSONDecodeError:
                logger.error("The secret content is not a valid JSON.")
                raise
        if not isinstance(current_value, dict):
            raise TypeError(
                "The current value of the secret must be a dict. Got {}".format(
                    type(current_value)
                )
            )

        if step == "createSecret":
            create_secret(lambda_session, secret_arn, token, current_value)

        elif step == "setSecret":
            pass

        elif step == "testSecret":
            pass

        elif step == "finishSecret":
            finish_secret(lambda_session, secret_arn, token)

        else:
            raise ValueError("Invalid step parameter")


def create_secret(lambda_session: Session, arn, token, current_value: dict):
    """
    Creates the new secret for vcluster and stores in Secret with AWSPENDING stage.
    First, we check that there are no AWSPENDING secret value already in place.
    """
    client = lambda_session.client("secretsmanager")
    try:
        client.get_secret_value(
            SecretId=arn, VersionId=token, VersionStage="AWSPENDING"
        )
        logger.warning(
            "createSecret: AWSPENDING already set for secret for {} - {}".format(
                arn, token
            )
        )
    except:
        logger.debug("No AWSPENDING secret already set. Clear to proceed.")
        try:
            new_secret_value: dict = new_gateway_vcluster_secret_value(
                current_value, lambda_session
            )
            client.put_secret_value(
                SecretId=arn,
                ClientRequestToken=token,
                SecretString=json.dumps(new_secret_value)
                if not isinstance(new_secret_value, str)
                else new_secret_value,
                VersionStages=["AWSPENDING"],
            )
            logger.info(
                "createSecret: Successfully put AWSPENDING stage secret for ARN %s and version %s."
                % (arn, token)
            )
        except client.exceptions.ResourceExistsException as error:
            logger.exception(error)
            logger.error(
                "A previous secret execution used this ClientToken. Must re-rotate the secret afterwards."
            )
        except Exception as error:
            logger.exception(error)
            logger.error(
                "createSecret: Failed to create a new secret for ARN %s and version %s."
                % (arn, token)
            )
            raise


def finish_secret(lambda_session: Session, arn, token):
    """
    Finalizes/Promotes the rotation process by marking the secret version passed in as the AWSCURRENT secret
    from AWSPENDING stage
    """
    # First describe the secret to get the current version
    service_client = lambda_session.client("secretsmanager")
    metadata = service_client.describe_secret(SecretId=arn)
    current_version = None
    for version in metadata["VersionIdsToStages"]:
        if "AWSCURRENT" in metadata["VersionIdsToStages"][version]:
            if version == token:
                # The correct version is already marked as current, return
                logger.info(
                    "finishSecret: Version %s already marked as AWSCURRENT for %s"
                    % (version, arn)
                )
                return
            current_version = version
            break

    # Finalize by staging the secret version current
    service_client.update_secret_version_stage(
        SecretId=arn,
        VersionStage="AWSCURRENT",
        MoveToVersionId=token,
        RemoveFromVersionId=current_version,
    )
    logger.info(
        "finishSecret: Successfully set AWSCURRENT stage to version %s for secret %s."
        % (token, arn)
    )



#   SPDX-License-Identifier: Apache-2.0
#   Copyright 2023 John Mille <john@ews-network.net>

"""Wrapper to configure mappings for a given tenant"""

from __future__ import annotations

import json
from dataclasses import asdict

from cdk_proxy_api_client.interceptors import Interceptors
from cdk_proxy_api_client.plugins import Plugins
from cdk_proxy_api_client.proxy_api import ProxyClient
from dacite import from_dict

from cdk_gw_tools.specs.interceptors_config import *


def set_update_for_interceptor_context(
    interceptors_client: Interceptors,
    plugin_class: str,
    interceptor_context: InterceptorContext,
    interceptor_name: str,
    vcluster_name: str = None,
):
    """Sets the interceptor for a given Interceptor Context"""
    if interceptor_context.definition:
        plugin_config: dict = asdict(interceptor_context.definition)
        plugin_config["pluginClass"] = plugin_class
        interceptors_client.update_interceptor(
            interceptor_name=interceptor_name,
            interceptor_config=plugin_config,
            vcluster_name=vcluster_name,
        )
    if interceptor_context.groups:
        for group_name, group_config in interceptor_context.groups.items():
            plugin_config: dict = asdict(group_config)
            plugin_config["pluginClass"] = plugin_class
            interceptors_client.update_interceptor(
                interceptor_name,
                interceptor_config=plugin_config,
                group_name=group_name,
                vcluster_name=vcluster_name,
            )
    if interceptor_context.usernames:
        for user_name, user_config in interceptor_context.usernames.items():
            plugin_config: dict = asdict(user_config)
            plugin_config["pluginClass"] = plugin_class
            interceptors_client.update_interceptor(
                interceptor_name,
                interceptor_config=plugin_config,
                username=user_name,
                vcluster_name=vcluster_name,
            )


def set_update_interceptors(proxy: ProxyClient, interceptors: dict):
    plugins_client = Plugins(proxy)
    interceptors_client = Interceptors(proxy)
    plugins_list = plugins_client.list_all_plugins(as_list=True)
    gw_interceptors_config: GwInterceptorsConfig = from_dict(
        data_class=GwInterceptorsConfig, data=interceptors
    )
    for (
        interceptor_name,
        interceptor_config,
    ) in gw_interceptors_config.interceptors.items():
        plugin_class = interceptor_config.pluginClass
        if plugin_class not in plugins_list:
            raise ValueError(
                "Interceptor %s does not exist. Available ones"
                % (interceptor_config.pluginClass,),
                plugins_list,
            )
        if interceptor_config.gateway:
            plugin_config: dict = asdict(interceptor_config.gateway)
            plugin_config["pluginClass"] = plugin_class
            interceptors_client.update_interceptor(
                interceptor_name=interceptor_name,
                interceptor_config=plugin_config,
                is_global=True,
            )
        if interceptor_config.passthrough:
            set_update_for_interceptor_context(
                interceptors_client,
                plugin_class,
                interceptor_config.passthrough,
                interceptor_name,
            )
        if interceptor_config.vclusters:
            for vcluster_name, vcluster_context in interceptor_config.vclusters.items():
                set_update_for_interceptor_context(
                    interceptors_client,
                    plugin_class,
                    vcluster_context,
                    interceptor_name,
                    vcluster_name=vcluster_name,
                )



# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 John Mille <john@ews-network.net>

from __future__ import annotations

from argparse import ArgumentParser

from cdk_gw_tools.cli.vclusters_parser.concentration_rules import (
    set_vcluster_concentration_actions,
)

VCLUSTER_PARSER = ArgumentParser(add_help=False)
VCLUSTER_PARSER.add_argument(
    "--vcluster-name",
    dest="vcluster_name",
    required=True,
    help="Name of the vcluster to make operations for",
)


def set_vclusters_actions_parsers(vclusters_subparsers):
    """Creates all the parser and subparsers for vClusters API endpoints"""

    # List all
    list_parser = vclusters_subparsers.add_parser(
        name="list", help="List all vClusters"
    )

    # Auth / Token
    tenant_token = vclusters_subparsers.add_parser(
        name="auth",
        help="Create a new JWT for vCluster/Username",
        parents=[VCLUSTER_PARSER],
    )
    tenant_token_subparsers = tenant_token.add_subparsers(dest="sub_action")
    tenant_token_create_parser = tenant_token_subparsers.add_parser(name="create")
    tenant_token_create_parser.add_argument(
        "--lifetime-in-seconds",
        dest="token_lifetime_in_seconds",
        type=int,
        help="Token lifetime in seconds. Sets expiry. Defaults 1 day (86400)",
        default=86400,
    )
    tenant_token_create_parser.add_argument(
        "--username",
        type=str,
        help="Sets the username to use for sasl.username to connect to the virtual cluster",
        required=False,
    )
    tenant_token_create_parser.add_argument(
        "--as-kafka-config",
        action="store_true",
        help="Returns the kafka config file",
        default=False,
    )
    set_vcluster_mappings_actions(vclusters_subparsers)
    set_vcluster_concentration_actions(vclusters_subparsers)


def set_vcluster_mappings_actions(vclusters_subparsers):
    # Mappings
    mappings_parser = vclusters_subparsers.add_parser(
        name="mappings",
        help="Manages vCluster mappings",
        parents=[VCLUSTER_PARSER],
    )
    mappings_subparsers = mappings_parser.add_subparsers(dest="sub_action")
    list_parser = mappings_subparsers.add_parser(
        name="list",
        help="List vCluster mappings",
    )
    list_parser.add_argument(
        "--no-concentrated",
        action="store_true",
        help="Excludes concentrated topics from list",
        dest="no_concentrated",
        required=False,
    )
    list_parser.add_argument(
        "--mapped-only",
        action="store_true",
        help="Excludes concentrated topics & topics starting with the vcluster name",
        dest="mapped_only",
        required=False,
    )
    list_parser.add_argument(
        "--as-import-config",
        action="store_true",
        help="Returns the equivalent output to use with import-from-config",
    )
    create_parser = mappings_subparsers.add_parser(
        name="create",
        help="Create a new vCluster mapping",
    )
    create_parser.add_argument("--logical-topic-name", type=str, required=True)
    create_parser.add_argument("--physical-topic-name", type=str, required=True)
    create_parser.add_argument(
        "--read-only",
        required=False,
        default=False,
        dest="ReadOnly",
        action="store_true",
        help="Creates mapping in ReadOnly (defaults to Read-Write)",
    )
    create_parser.add_argument(
        "--cluster-id",
        type=str,
        required=False,
        help="Creates a mapping for merged cluster",
    )
    # Delete action parsers
    delete_all_mappings_parser = mappings_subparsers.add_parser(
        name="delete-all-mappings",
        help="Delete all topics mappings for a given vCluster",
        parents=[],
    )
    delete_topic_mapping_parser = mappings_subparsers.add_parser(
        name="delete",
        help="Delete a topic mapping for a given vCluster",
        parents=[],
    )
    delete_topic_mapping_parser.add_argument(
        "--logical-topic-name",
        dest="logicalTopicName",
        required=True,
        help="Topic name as seen in the vCluster.",
    )

    # Custom cli_actions
    import_from_vclusters_parser = mappings_subparsers.add_parser(
        name="import-from-vclusters-config",
        help="Create topic mappings from existing vclusters",
    )
    import_from_vclusters_parser.add_argument(
        "-f",
        "--import-config-file",
        dest="import_config_file",
        help="Path to the mappings import file",
        required=True,
        type=str,
    )
    import_from_vcluster_parser = mappings_subparsers.add_parser(
        name="import-from-vcluster",
        help="Import all topics from a existing vCluster",
        parents=[],
    )
    import_from_vcluster_parser.add_argument(
        "--src",
        "--source-vcluster",
        dest="source_vcluster",
        help="Name of the source vCluster to import the mappings from",
        type=str,
        required=True,
    )



# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 John Mille <john@ews-network.net>

from __future__ import annotations

import json
import logging

import yaml
from cdk_proxy_api_client.client_wrapper import ApiClient
from cdk_proxy_api_client.proxy_api import ProxyClient
from yaml import Dumper

from cdk_gw_tools.cli.main_parser import set_parser
from cdk_gw_tools.cli_actions.auth import auth_actions
from cdk_gw_tools.cli_actions.interceptors import interceptors_actions
from cdk_gw_tools.cli_actions.plugins import plugins_actions
from cdk_gw_tools.cli_actions.user_mappings import user_mappings_actions
from cdk_gw_tools.cli_actions.vclusters import vclusters_actions
from cdk_gw_tools.cli_tools.import_from_config import import_client
from cdk_gw_tools.common.logging import LOG


def main():
    _PARSER = set_parser()
    _args = _PARSER.parse_args()
    if hasattr(_args, "loglevel") and _args.loglevel:
        valid_levels = [
            "FATAL",
            "CRITICAL",
            "ERROR",
            "WARNING",
            "WARN",
            "INFO",
            "DEBUG",
            "INFO",
        ]
        if _args.loglevel.upper() in valid_levels:
            LOG.setLevel(logging.getLevelName(_args.loglevel.upper()))
            LOG.handlers[0].setLevel(logging.getLevelName(_args.loglevel.upper()))
        else:
            print(
                f"Log level value {_args.loglevel} is invalid. Must me one of {valid_levels}"
            )
    _vars = vars(_args)
    if _args.url:
        if not _args.username or not _args.password:
            raise Exception(
                "If you specify URL, you must specify username and password too"
            )
        _client = ApiClient(
            username=_vars.pop("username"),
            password=_vars.pop("password"),
            url=_vars.pop("url"),
        )
    elif _args.profile_name:
        _client = import_client(_args.config_file, _args.profile_name)
    else:
        raise Exception(
            "You must either set --profile-name (possibly -c) or define --url, --username and --password"
        )
    _category = _vars.pop("category")
    _action = _vars.pop("action")
    _proxy = ProxyClient(_client)

    _categories_mappings: dict = {
        "vclusters": vclusters_actions,
        "plugins": plugins_actions,
        "user-mappings": user_mappings_actions,
        "interceptors": interceptors_actions,
        "auth": auth_actions,
    }
    dest_function = _categories_mappings[_category]
    response = dest_function(_proxy, _action, **_vars)
    if not response:
        return
    if isinstance(response, str):
        print(response)
        return
    try:
        if _args.output_format == "json":
            print(json.dumps(response, indent=2))
        else:
            print(yaml.dump(response, Dumper=Dumper))
    except Exception as error:
        print(error)
        print(response)



#   SPDX-License-Identifier: Apache-2.0
#   Copyright 2023 John Mille <john@ews-network.net>

from __future__ import annotations

import sys

try:
    from yaml import Dumper
except ImportError:
    from yaml import CDumper as Dumper

from cdk_gw_tools.cli_actions import main

if __name__ == "__main__":
    sys.exit(main())



#   SPDX-License-Identifier: Apache-2.0
#   Copyright 2023 John Mille <john@ews-network.net>

from __future__ import annotations

from typing import Union

import yaml

try:
    from yaml import Loader
except ImportError:
    from yaml import CLoader as Loader

import json
from os import path


def load_config_file(file_path: str) -> Union[dict, list]:
    """Simple function to load a file and return dict/list, the best effort to use YAML or JSON. If neither, fails."""
    with open(path.abspath(file_path)) as file_fd:
        content = file_fd.read()

    try:
        return yaml.load(content, Loader=Loader)
    except yaml.YAMLError:
        return json.loads(content)
    except json.JSONDecodeError:
        raise ValueError(f"The content of {file_path} is neither a valid YAML or JSON")



# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 John Mille <john@ews-network.net>

from __future__ import annotations

from argparse import ArgumentParser

VCLUSTER_PARSER = ArgumentParser(add_help=False)
VCLUSTER_PARSER.add_argument(
    "--vcluster-name",
    dest="vcluster_name",
    required=False,
    help="Name of the vCluster to make operations for",
)

USERNAME_PARSER = ArgumentParser(add_help=False)
USERNAME_PARSER.add_argument(
    "--username",
    dest="username",
    required=False,
    help="Specify username to get interceptors for",
)
GROUP_PARSER = ArgumentParser(add_help=False)
GROUP_PARSER.add_argument(
    "--group",
    dest="group",
    required=False,
    help="Specify username to get interceptors for",
)


def set_interceptors_actions_parsers(interceptors_subparsers):
    """Creates all the parser and subparsers for GW API endpoints"""
    list_all_parser = interceptors_subparsers.add_parser(
        name="list-all",
        help="List all the interceptors on GW for all contexts",
    )

    list_parser = interceptors_subparsers.add_parser(
        name="list",
        help="List interceptors",
        parents=[USERNAME_PARSER, VCLUSTER_PARSER, GROUP_PARSER],
    )
    list_parser.add_argument(
        "--global",
        action="store_true",
        help="List global interceptors.",
        dest="is_global",
    )

    create_parser = interceptors_subparsers.add_parser(
        name="create-update",
        help="Create or update a new vCluster mapping",
        parents=[USERNAME_PARSER, VCLUSTER_PARSER, GROUP_PARSER],
    )
    create_parser.add_argument(
        "--interceptor-name",
        dest="interceptor_name",
        required=True,
        help="Name of the interceptor to create or update",
    )
    create_parser.add_argument(
        "--global",
        action="store_true",
        help="Create or update a global interceptor",
        dest="is_global",
    )
    create_parser.add_argument(
        "--config",
        type=str,
        required=True,
        help="Path to config (YAML/JSON) to use for interceptor",
        dest="config_file_path",
    )

    # Delete action parsers
    delete_interceptor_mapping_parser = interceptors_subparsers.add_parser(
        name="delete",
        help="Delete interceptor",
        parents=[USERNAME_PARSER, VCLUSTER_PARSER, GROUP_PARSER],
    )
    delete_interceptor_mapping_parser.add_argument(
        "--interceptor-name",
        dest="interceptor_name",
        required=True,
        help="Interceptor name to delete as seen in the vCluster.",
    )
    delete_interceptor_mapping_parser.add_argument(
        "--global",
        action="store_true",
        help="Delete global interceptor",
        dest="is_global",
    )
    import_from_config_parser = interceptors_subparsers.add_parser(
        name="import-from-config",
        help="Import interceptors from config file",
    )
    import_from_config_parser.add_argument(
        "-f",
        "--config-file",
        dest="config_file_path",
        type=str,
        help="Path to the config file (YAML)",
        required=True,
    )



# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 John Mille <john@ews-network.net>

"""Manages the vclusters <action> of the CLI"""

from __future__ import annotations

from cdk_proxy_api_client.proxy_api import ProxyClient
from cdk_proxy_api_client.vclusters import VirtualClusters

from cdk_gw_tools.cli_actions.auth import auth_actions
from cdk_gw_tools.cli_actions.common import format_return
from cdk_gw_tools.cli_actions.interceptors import interceptors_actions
from cdk_gw_tools.cli_actions.user_mappings import user_mappings_actions
from cdk_gw_tools.cli_actions.vclusters.topic_mappings import vcluster_mappings_actions


def concentration_rules_actions(vclusters: VirtualClusters, action: str, **kwargs):
    if action == "list":
        req = vclusters.get_concentration_rules(
            vcluster_name=kwargs.get("vcluster_name")
        )
    elif action == "create":
        req = vclusters.create_concentration_rule(
            kwargs.get("vcluster_name"),
            kwargs.get("pattern"),
            delete_topic_name=kwargs.get("deleteTopicName"),
            compact_topic_name=kwargs.get(
                "compactTopicName", f"{kwargs.get('deleteTopicName')}_compact"
            ),
            compact_delete_topic_name=kwargs.get(
                "deleteCompactTopicName",
                f"{kwargs.get('deleteTopicName')}_delete_compact",
            ),
            cluster_id=kwargs.get("cluster_id"),
        )
    elif action == "delete":
        req = vclusters.delete_concentration_rule(
            vcluster_name=kwargs.get("vcluster_name"),
            pattern=kwargs.get("pattern"),
        )
    else:
        raise NotImplementedError(f"Action {action} not yet implemented.")
    return req


@format_return
def vclusters_actions(proxy: ProxyClient, action: str, **kwargs):
    """Manages execution of vClusters"""
    vclusters = VirtualClusters(proxy)
    if action == "list":
        req = vclusters.list_vclusters(as_list=True)
    elif action == "mappings":
        req = vcluster_mappings_actions(
            proxy, vclusters, kwargs.pop("sub_action"), **kwargs
        )
    elif action == "concentration-rules":
        req = concentration_rules_actions(vclusters, kwargs.pop("sub_action"), **kwargs)
    elif action == "interceptors":
        req = interceptors_actions(proxy, kwargs.pop("sub_action"), **kwargs)
    elif action == "user-mappings":
        req = user_mappings_actions(proxy, kwargs.pop("sub_action"), **kwargs)
    else:
        raise NotImplementedError(f"Action {action} not yet implemented.")
    return req



# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 John Mille <john@ews-network.net>

from __future__ import annotations

from argparse import ArgumentParser

VCLUSTER_PARSER = ArgumentParser(add_help=False)
VCLUSTER_PARSER.add_argument(
    "--vcluster-name",
    dest="vcluster_name",
    required=True,
    help="Name of the vcluster to make operations for",
)


def set_vcluster_concentration_actions(vclusters_subparsers):
    # Concentration rules
    mappings_parser = vclusters_subparsers.add_parser(
        name="concentration-rules",
        help="Manages vCluster concentration rules",
        parents=[VCLUSTER_PARSER],
    )
    concentration_subparsers = mappings_parser.add_subparsers(dest="sub_action")

    list_parser = concentration_subparsers.add_parser(
        name="list",
        help="List vCluster concentration rules",
    )

    create_parser = concentration_subparsers.add_parser(
        name="create",
        help="Create a new vCluster mapping",
    )
    create_parser.add_argument("--pattern", type=str, required=True, dest="pattern")
    create_parser.add_argument(
        "--delete-topic-name", type=str, required=True, dest="deleteTopicName"
    )
    create_parser.add_argument(
        "--compact-topic-name", type=str, required=False, dest="compactTopicName"
    )
    create_parser.add_argument(
        "--delete-compact-topic-name",
        type=str,
        required=False,
        dest="deleteCompactTopicName",
    )
    create_parser.add_argument(
        "--cluster-id",
        type=str,
        required=False,
        help="Creates a mapping for merged cluster",
        dest="clusterId",
    )

    delete_topic_mapping_parser = concentration_subparsers.add_parser(
        name="delete",
        help="Delete a topic mapping for a given vCluster",
        parents=[],
    )
    delete_topic_mapping_parser.add_argument(
        "--pattern",
        dest="pattern",
        required=True,
        help="Pattern concentration rule to remove",
    )



# generated by datamodel-codegen:
#   filename:  interceptors-config.spec.json

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, Optional


@dataclass
class InterceptorDefinition:
    priority: int
    config: Dict[str, Any]


TargetIdentity = Dict[str, InterceptorDefinition]


@dataclass
class InterceptorContext:
    definition: Optional[InterceptorDefinition] = None
    """
    Override definition for all entities within the context.
    """
    groups: Optional[TargetIdentity] = None
    """
    Override definition to apply to groups within the context.
    """
    usernames: Optional[TargetIdentity] = None
    """
    Override definition to apply to usernames within the context.
    """


@dataclass
class Interceptor:
    """
    Allows to define the same interceptor and its overrides. Global can only be defined once, an
    """

    pluginClass: Optional[str] = None
    """
    The plugin class to use for this interceptor. It has to be unique per name.
    """
    gateway: Optional[InterceptorDefinition] = None
    """
    Global interceptor definition
    """
    passthrough: Optional[InterceptorContext] = None
    """
    Interceptor override for GW passthrough.
    """
    vclusters: Optional[Dict[str, InterceptorContext]] = None
    """
    Configuration override for vClusters
    """


@dataclass
class GwInterceptorsConfig:
    """
    Specification for GW Interceptors config file
    """

    interceptors: Optional[Dict[str, Interceptor]] = None



#  SPDX-License-Identifier: MPL-2.0
#  Copyright 2020-2022 John Mille <john@compose-x.io>


from __future__ import annotations

import logging as logthings
import sys


class MyFormatter(logthings.Formatter):
    default_format = "%(asctime)s [%(levelname)8s] %(message)s"
    debug_format = "%(asctime)s [%(levelname)8s] (%(filename)s.%(lineno)d , %(funcName)s,) %(message)s"
    date_format = "%Y-%m-%d %H:%M:%S"

    def format(self, record) -> str:
        if record.levelno == logthings.DEBUG:
            formatter = logthings.Formatter(self.debug_format, self.date_format)
        else:
            formatter = logthings.Formatter(self.default_format, self.date_format)
        return formatter.format(record)


class InfoFilter(logthings.Filter):
    """Inspired from https://stackoverflow.com/a/16066513"""

    def filter(self, rec):
        return rec.levelno in (logthings.DEBUG, logthings.INFO)


class ErrorFilter(logthings.Filter):
    """Inspired from https://stackoverflow.com/a/16066513"""

    def filter(self, rec):
        return rec.levelno not in (logthings.DEBUG, logthings.INFO)


def setup_logging():
    """ """
    root_logger = logthings.getLogger()
    for h in root_logger.handlers:
        root_logger.removeHandler(h)

    app_logger = logthings.getLogger("cdk-proxy-cli")

    for h in app_logger.handlers:
        root_logger.removeHandler(h)

    stdout_handler = logthings.StreamHandler(sys.stdout)
    stdout_handler.setFormatter(MyFormatter())
    stdout_handler.setLevel(logthings.INFO)
    stdout_handler.addFilter(InfoFilter())

    stderr_handler = logthings.StreamHandler(sys.stderr)
    stderr_handler.setFormatter(MyFormatter())
    stderr_handler.setLevel(logthings.WARNING)
    stderr_handler.addFilter(ErrorFilter())

    app_logger.addHandler(stdout_handler)
    app_logger.addHandler(stderr_handler)
    app_logger.setLevel(logthings.INFO)
    return app_logger


LOG = setup_logging()



#   SPDX-License-Identifier: Apache-2.0
#   Copyright 2023 John Mille <john@ews-network.net>

"""Functions to help with the user-mappings management"""

from __future__ import annotations

import warnings
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from cdk_proxy_api_client.user_mappings import UserMappings
    from cdk_proxy_api_client.vclusters import VirtualClusters

from dacite import from_dict

from cdk_gw_tools.specs.user_mappings import DetailedIdentity, UserMappingsConfig


def create_export_config(
    vclusters: VirtualClusters, user_mappings: UserMappings
) -> UserMappingsConfig:
    """Lists all vClusters and returns the user mappings."""
    user_mappings_config: dict = {}
    mappings_config: dict = {"userMappings": user_mappings_config}
    _vclusters_list = vclusters.list_vclusters(as_list=True)
    for _vcluster_name in _vclusters_list["vclusters"]:
        _vcluster_mappings: list[dict] = []
        _vcluster_usernames: list[str] = user_mappings.list_mappings(
            vcluster_name=_vcluster_name
        ).json()
        for _username in _vcluster_usernames:
            _identity = user_mappings.get_user_mapping(
                username=_username, vcluster_name=_vcluster_name
            ).json()
            _vcluster_mappings.append(_identity)
        user_mappings_config[_vcluster_name] = {"identities": _vcluster_mappings}
    passthrough_users = user_mappings.list_mappings().json()
    passthrough_identities: list[dict] = []
    for _username in passthrough_users:
        passthrough_identities.append(
            user_mappings.get_user_mapping(username=_username).json()
        )
    user_mappings_config["passthrough"] = {"identities": passthrough_identities}
    return from_dict(UserMappingsConfig, mappings_config)


def validate_identities_are_unique(config: dict) -> UserMappingsConfig:
    mappings_config = from_dict(UserMappingsConfig, config)
    identities_recorded: list = []
    for definition in mappings_config.userMappings.values():
        for identity in definition.identities:
            if identity in identities_recorded:
                raise ValueError(f"Identity {identity} is defined multiple times.")
            identities_recorded.append(identity)
    return mappings_config


def handle_detailed_identity_input(
    user_mappings: UserMappings,
    existing_identities: list[tuple[str, str]],
    identity: DetailedIdentity,
    vcluster_name: str = None,
):
    """
    Checks if the username,principal combo already does exist. If it does,
    simply update (PUT) to update the mapping definition.
    If the combo doesn't exist, we evaluate against the existing pairs.
    Otherwise, we evaluate against each existing username,principal pair.
    If the username matches, but the principal doesn't, we update the principal, as this is an editable field
    If the username doesn't match, but the principal does, we delete the mapping altogether, and create it again, with
    the new username.
    If however none of these conditions are true, this is a new user-mapping and we create it.
    """
    to_create: tuple[str, str] = (
        identity.username,
        identity.principal,
    )
    if to_create in existing_identities:
        user_mappings.update_mapping(
            username=identity.username,
            principal=identity.principal,
            groups=identity.groups,
            vcluster_name=vcluster_name,
        )
        return
    for _existing_username, _existing_principal in existing_identities:
        if (
            identity.username == _existing_username
            and identity.principal != _existing_principal
        ):
            user_mappings.update_mapping(
                username=identity.username,
                principal=identity.principal,
                groups=identity.groups,
                vcluster_name=vcluster_name,
            )
            break
        elif (
            identity.username != _existing_username
            and identity.principal == _existing_principal
        ):
            user_mappings.delete_mapping(_existing_username, vcluster_name)
            user_mappings.create_mapping(
                username=identity.username,
                principal=identity.principal,
                groups=identity.groups,
                vcluster_name=vcluster_name,
            )
            break
    else:
        user_mappings.create_mapping(
            username=identity.username,
            principal=identity.principal,
            groups=identity.groups,
            vcluster_name=vcluster_name,
        )


def check_create_username_identity(
    user_mappings: UserMappings,
    identities_to_create: list[DetailedIdentity | str],
    existing_mappings: list[dict],
    vcluster_name: str | None = None,
) -> None:
    """
    Function to check and create the user mapping.
    Now (3.x+) that the user-mapping details has more than just the username, and previously the `username` was the
    `principal`, we check against the existence of both, just in case.
    """
    existing_identities: list[tuple[str, str]] = [
        (_mapping.get("username"), _mapping.get("principal"))
        for _mapping in existing_mappings
    ]
    for identity in identities_to_create:
        if isinstance(identity, str):
            warnings.warn(
                "For retro-compatibility, `strings` in the list of identities "
                "is maintained. It will be removed in a future version."
            )
            to_create = (
                identity,
                identity,
            )
            if to_create in existing_identities:
                user_mappings.update_mapping(
                    username=identity, principal=identity, vcluster_name=vcluster_name
                )
            else:
                user_mappings.create_mapping(
                    username=identity, principal=identity, vcluster_name=vcluster_name
                )
        elif isinstance(identity, DetailedIdentity):
            handle_detailed_identity_input(
                user_mappings,
                existing_identities,
                identity,
                vcluster_name=vcluster_name,
            )
        else:
            raise TypeError(
                f"Identity {identity} is not a string or a dict.", type(identity)
            )


def cleanup_undefined_identities(
    user_mappings: UserMappings,
    usernames_to_create: list[str],
    vcluster_name: str | None = None,
):
    """
    If a username is in existing identities but not in the list coming from the config file, delete
    the mapping.
    This helps with ensuring we use the config file as the target.
    """
    existing_usernames: list[str] = user_mappings.list_mappings(vcluster_name).json()
    for username in existing_usernames:
        if username not in usernames_to_create:
            user_mappings.delete_mapping(username=username, vcluster_name=vcluster_name)


def import_mappings_from_file(
    user_mappings: UserMappings, config: dict, remove_unset: bool = False
) -> None:
    """From the config file, will iterate over each vcluster to create/update/delete user-mappings."""
    mappings_config = validate_identities_are_unique(config)

    for _vcluster_name, definition in mappings_config.userMappings.items():
        vcluster_name: str | None = (
            None if _vcluster_name == "passthrough" else _vcluster_name
        )
        _existing_mappings = user_mappings.list_mappings_detailed(vcluster_name)
        usernames_to_create: list[str] = [
            _id.username
            for _id in definition.identities
            if isinstance(_id, DetailedIdentity)
        ] + [_id for _id in definition.identities if isinstance(_id, str)]
        check_create_username_identity(
            user_mappings,
            definition.identities,
            _existing_mappings,
            vcluster_name=vcluster_name,
        )
        if remove_unset:
            cleanup_undefined_identities(
                user_mappings, usernames_to_create, vcluster_name
            )



# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 John Mille <john@ews-network.net>

from __future__ import annotations


def set_auth_actions(auth_subparsers):
    tenant_token_create_parser = auth_subparsers.add_parser(
        name="create", help="Create a new tenant token"
    )
    tenant_token_create_parser.add_argument(
        "--lifetime-in-seconds",
        dest="token_lifetime_in_seconds",
        type=int,
        help="Token lifetime in seconds. Sets expiry. Defaults 1 day (86400)",
        default=86400,
    )
    tenant_token_create_parser.add_argument(
        "--vcluster-name",
        help="Specify the vCluster this user token is for",
        required=False,
    )
    tenant_token_create_parser.add_argument(
        "--username",
        type=str,
        help="Sets the username to use for sasl.username to connect to the virtual cluster",
        required=False,
    )
    tenant_token_create_parser.add_argument(
        "--as-kafka-config",
        action="store_true",
        help="Returns the kafka config file",
        default=False,
    )



# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 John Mille <john@ews-network.net>

from __future__ import annotations

from argparse import ArgumentParser

VCLUSTER_PARSER = ArgumentParser(add_help=False)
VCLUSTER_PARSER.add_argument(
    "--vcluster-name",
    dest="vcluster_name",
    required=False,
    help="Name of the vCluster to make operations for",
)


def set_user_mappings_actions_parsers(user_mappings_sub_parsers):
    list_parser = user_mappings_sub_parsers.add_parser(
        name="list",
        help="List user mappings",
        parents=[VCLUSTER_PARSER],
    )
    list_parser.add_argument(
        "--detailed",
        required=False,
        help="Returns the full list of identities with details",
        action="store_true",
        default=False,
        dest="detailed",
    )
    describe_parser = user_mappings_sub_parsers.add_parser(
        name="describe",
        help="Describe user mapping",
        parents=[VCLUSTER_PARSER],
    )
    describe_parser.add_argument(
        "--username",
        required=True,
        help="Username of the identity",
    )
    create_parser = user_mappings_sub_parsers.add_parser(
        name="create",
        help="Create user mapping",
        parents=[VCLUSTER_PARSER],
    )
    create_parser.add_argument(
        "--group",
        required=False,
        action="append",
        help="Groups this user is to be added into",
        default=[],
    )

    create_parser.add_argument(
        "--username",
        required=True,
        help="Friendly username.",
    )
    create_parser.add_argument(
        "--principal",
        required=True,
        help="Value of the `principal` used for Kafka authentication. When using OAuth, that's usuall the claims.sub value.",
    )

    delete_parser = user_mappings_sub_parsers.add_parser(
        name="delete",
        help="Create user mapping for vCluster",
        parents=[VCLUSTER_PARSER],
    )
    delete_parser.add_argument(
        "--username",
        required=True,
        help="Value of the `sub` of the JWT token from your issuer",
    )


def set_gw_user_mappings_actions_parsers(gw_user_mappings_subparsers):
    import_parser = gw_user_mappings_subparsers.add_parser(
        name="import", help="Import user-mappings from config-file"
    )
    import_parser.add_argument(
        "-f",
        "--config-file",
        dest="import_file",
        type=str,
        help="Path to the import config file",
    )
    import_parser.add_argument(
        "--remove-unset",
        action="store_true",
        help="If true, remove mappings on GW not defined in the config file",
    )

    export_parser = gw_user_mappings_subparsers.add_parser(
        name="export", help="Export user-mappings from config-file"
    )
    export_parser.add_argument(
        "-f",
        "--config-file",
        dest="import_file",
        type=str,
        help="Path to the import config file",
        required=False,
    )



# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 John Mille <john@ews-network.net>

from __future__ import annotations

from dataclasses import asdict

from cdk_proxy_api_client.proxy_api import ProxyClient
from cdk_proxy_api_client.user_mappings import UserMappings
from cdk_proxy_api_client.vclusters import VirtualClusters
from compose_x_common.compose_x_common import keyisset

from cdk_gw_tools.cli_actions.common import format_return
from cdk_gw_tools.cli_tools import load_config_file
from cdk_gw_tools.cli_tools.user_mappings_tools import (
    create_export_config,
    import_mappings_from_file,
    validate_identities_are_unique,
)


@format_return
def user_mappings_actions(proxy: ProxyClient, action: str, **kwargs):
    """
    Handles user-mappings cli_actions
    """
    vclusters = VirtualClusters(proxy)
    user_mappings = UserMappings(proxy)
    if action == "list":
        if kwargs.get("detailed"):
            return user_mappings.list_mappings_detailed(
                vcluster_name=kwargs.get("vcluster_name", None)
            )
        return user_mappings.list_mappings(
            vcluster_name=kwargs.get("vcluster_name", None)
        )
    elif action == "describe":
        return user_mappings.get_user_mapping(
            username=kwargs.get("username"), vcluster_name=kwargs.get("vcluster_name")
        )
    elif action == "create":
        return user_mappings.create_mapping(
            kwargs.get("username"),
            principal=kwargs.get("principal"),
            groups=kwargs.get("groups"),
            vcluster_name=kwargs.get("vcluster_name"),
        )
    elif action == "delete":
        return user_mappings.delete_mapping(
            username=kwargs.get("username"),
            vcluster_name=kwargs.get("vcluster_name", None),
        )
    elif action == "export":
        req = create_export_config(vclusters, user_mappings)
        return asdict(req)
    elif action == "import":
        config = load_config_file(kwargs["import_file"])
        import_mappings_from_file(
            user_mappings, config, keyisset("remove_unset", kwargs)
        )
        return {"status": "imported"}
    elif action == "validate":
        config = load_config_file(kwargs["import_file"])
        validate_identities_are_unique(config)
        return {"status": "valid"}
    else:
        raise NotImplementedError(f"Action {action} is not implemented yet.")
